%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% the one class support vector machine solution path
%
% gyemin lee    2006/10/12
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [lambda alpha elbow] =	ocsvm_path(x, hkernel, kernelparam)
% function [lambda alpha elbow] = ocsvm_path(x, hkernel, kernelparam)
%	x				n x p feature vector
%	hkernel			kernel function handle
%	kernelparam		kernel parameter
%	btrace			boolean trace
%
if nargin < 2, hkernel = @knl_poly; end
if nargin < 3, kernelparam = 1; end
btrace = false;

EPS = 1e-10;
EPS2 = 1e-14;
% lambdamin = 1e-4/size(x,1);
lambdamin = EPS;

n = size(x,1);
Left = (1:n)';
Elbow = [];
Right = [];

nmoves = 5*n;				% maximum number of steps
lambda = zeros(nmoves,1);
alpha = zeros(n,nmoves);
stats = zeros(nmoves,3);
elbow = cell(nmoves,1);

K = hkernel(x, x, kernelparam);
K = (K+K')/2;

k = 1;		% step

[lambda0 Elbow0] = ocsvminit(K);
lambda(k) = lambda0 / n; %---
Elbow = Elbow0;
Left = setdiff(Left, Elbow);
alpha(:,k) = ones(n, 1) / n; %---
Kstar = K(Elbow, Elbow);
fl = K * alpha(:,k) / lambda(k);

movefrom = repmat(' ', 1, length(Elbow));
moveto = repmat('E', 1, length(Elbow));
obs = Elbow;

stats(k,:) = ocsvmstat(fl, Elbow);
elbow{k} = Elbow;

if (btrace)
	strtrace = ocsvmprint(k, obs, movefrom, moveto, lambda(k), stats(k,:));
end

while (k<nmoves && lambda(k)>lambdamin && ~isempty(Left))
	if (isempty(Elbow))
		% the elbow is empty. need to reinitialize.
		[lambda0 Elbow0] = ocsvminit(K(Left, Left));
        if (isempty(lambda0) || lambda0<0)
            break
        end
        if (lambda(k)<lambda0 / n)
            fprintf('WARNING: OC-SVM : lambda(k)<lambda0 \n');
            break
        end
		lambda(k+1) = lambda0 / n; %---
		Elbow = Left(Elbow0);
		Left = setdiff(Left, Elbow);
		alpha(:,k+1) = alpha(:,k);
		Kstar = K(Elbow, Elbow);
% 		fl = lambda(k) / lambda(k+1) * fl;
        fl = K * alpha(:,k+1) / lambda(k+1);
		
		movefrom = repmat(' ', 1, length(Elbow));
		moveto = repmat('E', 1, length(Elbow));
		obs = Elbow;
	else
		b = Kstar\ones(size(Kstar,1),1);
		
		hl = K(:,Elbow) * b;
		dl = fl - hl;
		
		% check for immobile margin
		immobile = sum(abs(dl))/n < EPS2;
		
		% check for exits from the elbow
		temp = ~(abs(b)<EPS2);
		lambdaleft = -ones(size(temp));
		lambdaright = -ones(size(temp));
        if sum(temp)
    		lambdaleft(temp) = (1/n-alpha(Elbow(temp),k)+lambda(k)*b(temp))./b(temp); %---
    		lambdaright(temp) = (-alpha(Elbow(temp),k)+lambda(k)*b(temp))./b(temp);
        end
% 		lambdaleft = zeros(size(temp));
% 		lambdaleft(temp) = (1/n-alpha(Elbow(temp),k)+lambda(k)*b(temp))./b(temp); %---
% 		lambdaleft(~temp) = -1;
% 		lambdaright = zeros(size(temp));
% 		lambdaright(temp) = (-alpha(Elbow(temp),k)+lambda(k)*b(temp))./b(temp);
% 		lambdaright(~temp) = -1;
		lambdarl = [lambdaright; lambdaleft];
		lambdaexit = max(lambdarl(lambdarl < lambda(k) - EPS2));
		if isempty(lambdaexit), lambdaexit = -1; end
		
		% when immobile, leave
% 		if (immobile && lambdaexit<lambdamin), break, end
		
		% check for entry into the elbow
        if (~immobile)
			temp = ~(abs(1-hl)<EPS2);
            temp(Elbow) = false;
            temp(obs) = false;
            lambdae = -ones(size(temp));
			lambdae(temp) = lambda(k).*dl(temp)./(1-hl(temp));
% 			lambdae(~temp) = -1;
			lambdaentry = max(lambdae(lambdae < lambda(k) - EPS2));
% 			lambdaentry = max(lambdae(lambdae < lambda(k)));
			if isempty(lambdaentry), lambdaentry = -1; end
		else
			lambdaentry = -1;
        end
		
		lambdamax = max(lambdaexit, lambdaentry);
		if (lambdamax<=0), break, end
		
		% update lambda, alpha
		lambda(k+1) = lambdamax;
        if (lambda(k+1)>lambda(k)), break, end
		alpha(:,k+1) = alpha(:,k);
        alpha(Right,k+1) = 0;
        alpha(Left,k+1) = 1/n;
		alpha(Elbow,k+1) = alpha(Elbow,k) - (lambda(k)-lambdamax)*b;
% 		fl = lambda(k) / lambda(k+1) * dl + hl;
		fl = K*alpha(:,k+1)/lambda(k+1);
        
		% update sets
        if (lambdaentry > lambdaexit)
% 			iin = find(lambdae==lambdaentry);
            iin = find(abs(lambdae-lambdaentry)<EPS2);
			obs = iin;
			
			tempe = ismember(iin, Left);
			movefrom = repmat('R', 1, length(iin));
			movefrom(tempe) = 'L';
			
			Left = setdiff(Left, iin);
			Right = setdiff(Right, iin);
			
			Elbow = [Elbow; iin];
			Kstar = K(Elbow, Elbow);
			moveto = repmat('E', 1, length(iin));
        else
		    tempr = abs(lambdaright - lambdaexit)<EPS2;
		    toright = Elbow(tempr);
		    templ = abs(lambdaleft - lambdaexit)<EPS2;
		    toleft = Elbow(templ);
		    temp = tempr | templ;
		    
		    Elbow = Elbow(~temp);
		    Right = [Right; toright];
		    moveto = repmat('R', 1, length(toright));
		    Left = [Left; toleft];
		    moveto = [moveto, repmat('L', 1, length(toleft))];
		
		    Kstar = K(Elbow,Elbow);
		    obs = [toright; toleft];
			movefrom = repmat('E', 1, length(moveto));
        end
	end
	
	k = k + 1;
	
	stats(k,:) = ocsvmstat(fl, Elbow);
	elbow{k} = Elbow;
	
	if (btrace)
		strtrace = [strtrace, ...
			ocsvmprint(k, obs, movefrom, moveto, lambda(k), stats(k,:))];
	end
end

% outputs
lambda = lambda(1:k);
alpha = alpha(:,1:k);
stats = stats(1:k,:);
elbow = elbow(1:k);

% feasiblity check
if ~isempty(find(alpha < -EPS*2 | alpha > 1/n + 2*EPS,1));
    fprintf('WARNING: OC-SVM : infeasible solution!!\n');
end
if k>=nmoves
    fprintf('WARNING: OC-SVM : max breakpoint num \n');
end

% print the result
if (btrace)
	fresult = fopen('ocsvm_path.txt', 'w');
	fwrite(fresult, strtrace);
	fclose(fresult);
%	strtrace
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [lambda Elbow] = ocsvminit(K)
%function [lambda Elbow] = ocsvminit(K)
%	ocsvm initialization
%
EPS2 = 1e-14;

gk = sum(K,2);
lambda = max(gk);
Elbow = find(lambda-gk<EPS2);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function stats = ocsvmstat(ff, Elbow)
%function stats = ocsvmstat(ff, Elbow)
%	ocsvm statistics
%	size of Elbow, number of error points, sum of margin errors
%
EPS = 1e-10;

stats = zeros(1,3);
stats(1) = length(Elbow);		% size of Elbow
stats(2) = sum(ff<1);   	% error
stats(3) = sum(1-ff(ff<1));		% margin


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function str = ocsvmprint(step, obs, movefrom, moveto, lambda, stats)
%function str = ocsvmprint(step, obs, movefrom, moveto, lambda, stats)
%	trace the ocsvm solution path
%
str = '';
for ii=1:length(obs)
	strtemp = sprintf('%3d: Obs %3d %c->%c lambda=%6.6f  Elbow=%3d  Error=%3d  Margin=%3.2f\n', ...
		step, obs(ii), movefrom(ii), moveto(ii), lambda, stats(1), stats(2), stats(3));
	str = [str, strtemp];
end
